"""
A module to aid working with histograms
"""

from __future__ import annotations

# ext
import numba
import numba.types
import numpy as np
from uproot_methods.classes import TH1


class CustomTH1(TH1.Methods, list):
    """A TH1 like skeleton object"""

    pass


class CustomTAxis:
    """A TAxis like object"""

    def __init__(self, edges: np.ndarray) -> None:
        self._fNbins = len(edges) - 1
        self._fXmin = edges[0]
        self._fXmax = edges[-1]
        self._fXbins = edges.astype(np.float64)


@numba.njit(numba.types.UniTuple(numba.float64[:], 2)(numba.float64[:], numba.float64[:]))
def prepare_padded(
    content: numpy.ndarray, errors: numpy.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Prepare arrays for saving to ROOT histogram with over/underflow

    This is accelerated with numba because it's 4x faster than
    non-jit'd version.

    Paramters
    ---------
    content : :py:obj:`numpy.ndarray`
       the bin contents
    error : :py:obj:`numpy.ndarray`
       the error on the bin content (the square-root of the variances)

    Returns
    -------
    :py:obj:`numpy.ndarray`
       the padded content
    :py:obj:`numpy.ndarray`
       the padded sumw2
    """
    content_padded = np.empty(len(content) + 2, dtype=content.dtype)
    content_padded[1:-1] = content
    content_padded[0] = 0.0
    content_padded[-1] = 0.0
    sumw2_padded = np.empty(len(content) + 2, dtype=np.float64)
    sumw2_padded[1:-1] = error ** 2
    sumw2_padded[0] = 0.0
    sumw2_padded[-1] = 0.0
    return content_padded, sumw2_padded


def from_pygram11(
    content: numpy.ndarray, error: numpy.ndarray, bins: numpy.ndarray, title: str = "none"
) -> CustomTH1:
    """create a TH1-like object built from pygram11 output

    Parameters
    ----------
    content : :py:obj:`numpy.ndarray`
       the bin contents
    error : :py:obj:`numpy.ndarray`
       the error on the bin content (the square-root of the variances)
    bins : :py:obj:`numpy.ndarray`
       the binning definition
    title : str
       title the histogram

    Returns
    -------
    :obj:`CustomTH1`
       the ROOT like histogram structure
    """

    output = CustomTH1.__new__(CustomTH1)
    if content.dtype == np.float32:
        output._classname = "TH1F"
    elif content.dtype == np.float64:
        output._classname = "TH1D"
    output._fXaxis = CustomTAxis(bins)
    output._fEntries = content.sum()
    output._fTitle = name

    content_padded, output._fSumw2 = prepare_padded(content, error)
    output.extend(content_padded)

    return output



"""
Module for art from TRExFitter
"""

from __future__ import annotations

# stdlib
import logging
import os
import re
import subprocess
from dataclasses import dataclass, field
from pathlib import PosixPath

# externals
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import uproot
import yaml

# tdub
import tdub.utils
from tdub._art import draw_atlas_label, setup_style, var_to_axis_meta

log = logging.getLogger(__name__)


@dataclass
class Sample:
    """Defines a physics sample

    Attributes
    ----------
    name : str
       name of the sample
    signature : str
       name of the sample on disk
    color : str
       matplotlib color
    tex : str
       LaTeX label
    """

    name: str = "none"
    signature: str = "none"
    color: str = "black"
    tex: str = "none"


@dataclass
class NuisPar:
    """Defines a nuisance parameter from TRExFitter

    Attributes
    ----------
    name : str
       name on disk
    mean : float
       mean value
    minus : float
       minus part of the error bar
    plus : float
       plus part of the error bar
    category : str
       systematic category
    title : str
       a title for the plots
    """

    name: str = "none"
    mean: float = 0
    minus: float = 0
    plus: float = 0
    category: str = "none"
    title: str = "none"


@dataclass
class Template:
    """Defines a template for generation by TREx

    Attributes
    ----------
    var : str
       variable (branch) from the tree
    region : List[str]
       list of regions where the template was calculated
    xmin : float
       axis range minimum
    xmax: float
       axis range maximum
    nbins : int
       number of bins
    use_region_binning : bool
       for wt-stat usage
    axis_title : str
       axis title for TRExFitter
    mpl_title : str
       axis title for matplotlib
    is_aux : bool
       for wt-stat usagse
    unit : str
       unit as a string (e.g. GeV)
    """

    var: str = ""
    regions: List[str] = field(default_factory=list)
    xmin: float = 0
    xmax: float = 0
    nbins: int = 0
    use_region_binning: bool = False
    axis_title: str = ""
    mpl_title: str = ""
    is_aux: bool = False
    unit: str = ""


@dataclass
class TRExHistogram:
    """Defines a histogram built from a TRExFitter output file

    Attributes
    ----------
    hfile : str
       the name of the TRExFitter histogram file holding the data
    region : str
       the region as defined in the TRExFitter configuration
    sample : Sample
       which physics sample
    unit : str
       a unit for plots (e.g. "GeV")
    mpl_title : str
       an axis title for matplotlib usage
    signature : str
       the actual name of the TH1 object in hte ROOT file
    postfit : bool
       whether or not the histogram is a postfit histogram
    """

    hfile: str
    region: str
    sample: Sample
    unit: str = ""
    mpl_title: str = ""
    signature: str = ""
    postfit: bool = False

    def __post_init__(self):
        pfp = "" if not self.postfit else "h_"
        pfs = "" if not self.postfit else "_postFit"
        regionsig = f"{self.region}_" if not self.postfit else ""
        self.signature = f"{pfp}{regionsig}{self.sample.signature}{pfs}"
        try:
            self.uproothist = uproot.open(self.hfile).get(self.signature)
            self.content = self.uproothist.values
            self.content[self.content < 0] = 1.0e-6
        except KeyError:
            self.uproothist = None
            self.content = None

    def __bool__(self):
        return self.uproothist is not None

    def __call__(self):
        return self.uproothist

    @property
    def sumw2(self) -> numpy.ndarray:
        """numpy.ndarray: the sum of weights squared in each bin"""
        return self.uproothist.variances

    @property
    def error(self) -> numpy.ndarray:
        """numpy.ndarray: the uncertainty in each bin (sqrt of sumw2)"""
        return np.sqrt(self.sumw2)

    @property
    def bins(self) -> numpy.ndarray:
        """numpy.ndarray: the bin edges"""
        return self.uproothist.edges

    @property
    def bin_centers(self) -> numpy.ndarray:
        """numpy.ndarray: the bin centers"""
        return tdub.utils.bin_centers(self.bins)

    @property
    def bin_width(self) -> numpy.ndarray:
        """numpy.ndarray: the bin widths"""
        return round(self.bins[-1] - self.bins[-2], 2)

    def has_uniform_bins(self) -> bool:
        """determines if the histogram has uniform bin widths

        Returns
        -------
        bool
           whether or not bin widthds are uniform.
        """
        diffs = np.ediff1d(self.bins)
        return np.allclose(diffs, diffs[0])


def _get_plot_samples() -> List[Sample]:
    plot_samples = [
        Sample(name="tW", signature="tW", color="#1f77b4", tex="$tW$"),
        Sample(name="ttbar", signature="ttbar", color="#d62728", tex="$t\\bar{t}$"),
        Sample(name="Diboson", signature="Diboson", color="#2ca02c", tex="Diboson"),
        Sample(name="Zjets", signature="Zjets", color="#ff7f0e", tex="$Z+$jets"),
        Sample(name="MCNP", signature="MCNP", color="#9467bd", tex="MCNP"),
    ]
    plot_samples.reverse()
    return plot_samples


_region_meta = var_to_axis_meta()


def draw_ratio_with_line(
    ax: matplotlib.axes.Axis,
    data: TRExHistogram,
    mc_sum: numpy.ndarray,
    mc_err: numpy.ndarray,
    yline: float = 1.0,
    autoxscale: bool = True,
) -> None:
    """ draw the ratio with a horizontal line on the axis """
    x1 = data.bins[0]
    x2 = data.bins[-1]
    err = np.sqrt(
        data.content / (mc_sum ** 2) + np.power(data.content * mc_err / (mc_sum ** 2), 2)
    )
    ax.plot([x1, x2], [yline, yline], color="gray", linestyle="solid", marker=None)
    ax.errorbar(data.bin_centers, data.content / mc_sum, yerr=err, fmt="ko", zorder=999)
    ax.set_ylabel("Data / MC")
    if autoxscale:
        ax.autoscale(enable=True, axis="x", tight=True)


def set_labels(ax: matplotlib.axes.Axes, histogram: TRExHistogram) -> None:
    """ define the axis labels """
    if histogram.has_uniform_bins():
        ylabel_suffix = f" / {histogram.bin_width} {histogram.unit}"
    else:
        ylabel_suffix = f" / bin"
    ax.set_ylabel(f"Events{ylabel_suffix}", horizontalalignment="right", y=1.0)


def shrink_pdf(file_path: str) -> None:
    """ shrink pdf file using ghostscript """
    command = (
        "gs -sDEVICE=pdfwrite -dCompatibilityLevel=1.4 "
        "-dNOPAUSE -dQUIET -dBATCH "
        "-sOutputFile=temp.pdf {in_file}"
    )
    log.debug(f"shrinking {file_path} via gs")
    in_file = PosixPath(file_path).absolute()
    in_name = os.fspath(in_file)
    proc = subprocess.Popen(command.format(in_file=in_file), shell=True)
    proc.wait()
    os.remove(in_file)
    os.rename("temp.pdf", in_name)


def stackem(
    args: argparse.Namespace,
    region: str,
    data: Sample,
    histograms: List[Sample],
    template_var: str,
    band: Optional[uproot_methods.classes.TGraphAasymmErrors] = None,
    figsize: Optional[Tuple[float, float]] = None,
) -> Tuple[matplotlib.figure.Figure, Tuple[matplotlib.axes.Axis, matplotlib.axes.Axis]]:
    """Create a stack plot

    Parameters
    ----------
    args : argparse.Namespace
       command line arguments
    region : str
       region from TRExFitter
    data : Sample
       data sample
    histograms : list(Sample)
       list of MC samples to stack
    template_var : str
       name of the template variable
    band : Optional[uproot_methods.classes.TGraphAsymmErrors]
       error band
    figsize : tuple(float, float), optional
       matplotlib figure size

    Returns
    -------
    fig : matplotlib.figure.Figure
       matplotlib fiture
    axes : tuple(matplotlib.axes.Axes, matplotlib.axes.Axes)
       matplotlib axes (main, ratio)
    """
    # fmt: off
    fig, (ax, axr) = plt.subplots(2, 1, sharex=True, figsize=figsize,
                                  gridspec_kw=dict(height_ratios=[3.25, 1], hspace=0.025))
    expected_sum = np.sum([h.content for h in histograms], axis=0)
    expected_err = np.sqrt(np.sum([h.sumw2 for h in histograms], axis=0))
    ax.hist([h.bin_centers for h in histograms], weights=[h.content for h in histograms],
            bins=histograms[0].bins, histtype="stepfilled", stacked=True,
            color=[h.sample.color for h in histograms], label=[h.sample.tex for h in histograms])
    ax.errorbar(data.bin_centers, data.content, yerr=data.error, fmt="ko", label="Data", zorder=999)
    set_labels(ax, histograms[0])

    if _region_meta[template_var].logy:
        ax.set_yscale("log")
        ax.set_ylim([ax.get_ylim()[0], ax.get_ylim()[1] * 1000])
    else:
        ax.set_ylim([ax.get_ylim()[0], ax.get_ylim()[1] * 1.5])

    if band is not None:
        yerrlo = np.hstack([band.yerrorslow, band.yerrorslow[-1]])
        yerrhi = np.hstack([band.yerrorshigh, band.yerrorshigh[-1]])
        expected_sum4band = np.hstack([expected_sum, expected_sum[-1]])

        if args.band_style == "hatch":
            ax.fill_between(x=data.bins,y1=(expected_sum4band - yerrlo), y2=(expected_sum4band + yerrhi),
                            step="post", facecolor="none", hatch="////", edgecolor="cornflowerblue",
                            linewidth=0.0, zorder=50, label="Syst. Unc.")
            axr.fill_between(x=data.bins, y1=1 - yerrlo / expected_sum4band, y2=1 + yerrhi / expected_sum4band,
                             step="post", facecolor="none", hatch="////", edgecolor="cornflowerblue",
                             linewidth=0.0, zorder=50, label="Syst. Unc.")
        elif args.band_style == "shade":
            axr.fill_between(x=data.bins, y1=1 - yerrlo / expected_sum4band, y2=1 + yerrhi / expected_sum4band,
                             step="post", facecolor=(0, 0, 0, 0.33), linewidth=0.0, zorder=50, label="Syst. Unc.")

    # fmt: on
    draw_ratio_with_line(axr, data, expected_sum, expected_err)
    axr.set_ylim([0.8, 1.2])
    axr.set_yticks([0.9, 1.0, 1.1])
    axis_title = "{}{}".format(data.mpl_title, "" if data.unit == "" else f" [{data.unit}]")
    axr.set_xlabel(axis_title, horizontalalignment="right", x=1.0)

    if band is not None and args.band_style == "shade":
        axr.legend(loc="upper center", frameon=True, fontsize=8)

    ax.legend(loc="upper right")
    handles, labels = ax.get_legend_handles_labels()
    handles.insert(0, handles.pop())
    labels.insert(0, labels.pop())
    ax.legend(handles, labels, loc="upper right", ncol=args.legend_ncol)

    raw_region = region.split("reg")[-1].split("_")[0]
    extra_line1 = f"$\\sqrt{{s}}$ = 13 TeV, $L = {args.lumi}$ fb$^{{-1}}$"
    extra_line2 = f"$pp\\rightarrow tW \\rightarrow e^{{\\pm}}\\mu^{{\\mp}} ({raw_region})$"
    extra_line3 = "Pre-fit"
    if histograms[0].postfit:
        extra_line3 = "Post-fit"
    draw_atlas_label(ax, extra_lines=[extra_line1, extra_line2, extra_line3])

    fig.subplots_adjust(left=0.115, bottom=0.115, right=0.965, top=0.95)
    return fig, (ax, axr)


def prefit_histograms(
    args: argparse.Namespace, fit_name: str, region: str, samples: List[Sample]
) -> Tuple[TRExHistogram, List[TRExHistogram], uproot_methods.classes.TGraphAsymmErrors]:
    """Prepare prefit histogram objects

    Parameters
    ---------
    args : argparse.Namespace
       command line arguments
    fit_name : str
       TRExFitter fit name
    region : str
       TRExFitter region
    samples : list(Sample)
       list of MC samples

    Returns
    -------
    data : TRExHistogram
       data histogram
    histograms : List[TRExHistogram]
       MC histograms
    band : uproot_methods.classes.TGraphAsymmErrors
       uncertainty band
    """
    hfile = f"{args.workspace}/Histograms/{fit_name}_{region}_histos.root"
    bfile = f"{args.workspace}/Histograms/{fit_name}_{region}_preFit.root"
    band = uproot.open(bfile).get("g_totErr")
    histograms = [TRExHistogram(hfile, region, sample) for sample in samples]
    data = TRExHistogram(hfile, region, Sample("Data", "Data", "", "Data"))
    return data, histograms, band


def postfit_histograms(
    args: argparse.Namespace, fit_name: str, region: str, samples: List[Samples]
) -> Tuple[TRExHistogram, List[TRExHistogram], uproot_methods.classes.TGraphAsymmErrors]:
    """Prepare postfit histogram objects

    Parameters
    ---------
    args : argparse.Namespace
       command line arguments
    fit_name : str
       TRExFitter fit name
    region : str
       TRExFitter region
    samples : List[Sample]
       list of MC samples

    Returns
    -------
    data : TRExHistogram
       data histogram
    histograms : List[TRExHistogram]
       MC histograms
    band : uproot_methods.classes.TGraphAsymmErrors
       uncertainty band

    """
    hfile = f"{args.workspace}/Histograms/{region}_postFit.root"
    bfile = f"{args.workspace}/Histograms/{region}_postFit.root"
    band = uproot.open(bfile).get("g_totErr_postFit")
    histograms = [TRExHistogram(hfile, region, sample, postfit=True) for sample in samples]
    return histograms, band


def split_region_str(region: str) -> Tuple[str, str]:
    splits = region.split("_")
    if len(splits) == 1:
        return (region, "bdt_response")
    else:
        return (splits[0], "_".join(splits[1:]))


def run_stacks(args: argparse.Namespace) -> None:
    """Given command line arguments generate stack plots

    Parameters
    ----------
    args : argparse.Namespace

    """
    setup_style()
    samples = _get_plot_samples()

    if args.out_dir is None:
        outd = f"{args.workspace}/MPL"
    else:
        outd = args.out_dir
    if outd != ".":
        PosixPath(outd).mkdir(parents=True, exist_ok=True)

    fit_name = PosixPath(args.workspace).stem
    hfiledir = PosixPath(f"{fit_name}/Histograms")
    regions = []
    if args.skip_regions is not None:
        skipregex = re.compile(args.skip_regions)
    else:
        skipregex = None
    for hfile in hfiledir.iterdir():
        if "_histos.root" in hfile.name:
            region = hfile.name.split("_histos.root")[0].split(f"{fit_name}_")[-1]
            if skipregex:
                if re.search(skipregex, region):
                    continue
            regions.append(region)

    for region in regions:
        raw_region, template_variable = split_region_str(region)
        data, histograms, band = prefit_histograms(args, fit_name, region, samples)
        data.unit = _region_meta[template_variable].unit
        data.mpl_title = _region_meta[template_variable].title
        fig, (ax, axr) = stackem(
            args, region, data, histograms, template_variable, band=band
        )
        out_name = f"{outd}/preFit_{region}.pdf"
        fig.savefig(out_name)
        plt.close(fig)
        log.info(f"Done with {region} prefit")

        if args.do_postfit:
            histograms, band = postfit_histograms(args, fit_name, region, samples)
            fig, (ax, axr) = stackem(
                args, region, data, histograms, template_variable, band=band
            )
            axr.set_ylim([0.925, 1.075])
            axr.set_yticks([0.95, 1.0, 1.05])
            out_name = f"{outd}/postFit_{region}.pdf"
            fig.savefig(out_name)
            plt.close(fig)
            log.info(f"Done with {region} postfit")


def get_blank_systematics(config_file: str) -> Tuple[List[NuisPar], Set[str]]:
    """Get list of NPs and categories from TRExFitter config file

    Parameters
    ----------
    config_file : str
        name of config file

    Returns
    -------
    nps : list(NuisPar)
        list of nuisance parameters
    categories: set(str)
        all categories that were found
    """
    trex_config = PosixPath(config_file)
    nps = []
    with trex_config.open("r") as f:
        config_blocks = f.read().split("\n\n")
        for block in config_blocks:
            if "Systematic:" in block:
                if block[0] == "%":
                    continue
                block = block.replace("\n  ", "\n")
                s = yaml.load(block, Loader=yaml.FullLoader)
                np = NuisPar(name=s["Systematic"], category=s["Category"], title=s["Title"])
                nps.append(np)
    categories = set()
    nps = {np.name: np for np in nps}
    for npn, np in nps.items():
        categories.add(np.category)
    return nps, categories


def draw_pulls(
    args: argparse.Namespace, nps: List[NuisPar]
) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    """Draw pulls from command line arguments and nuisance parameters

    Parameters
    ----------
    args : argparse.Namespace
       command line arguments
    nps : list
       list of nuisance parameters to draw

    Returns
    -------
    fig : matplotlib.figure.Figure
       matplotlib fiture
    ax : matplotlib.axes.Axes
       matplotlib axis
    """
    Y_OFFSET_TEXT = 0.095
    Y_OFFSET_TEXT_MEAN = 0.165
    X_OFFSET_TEXT = 0.035

    nnps = len(nps)
    xval = np.array([np.mean for np in nps])
    yval = np.array([(i + 1) for i in range(len(xval))])
    xerr_lo = np.array([np.minus for np in nps])
    xerr_hi = np.array([np.plus for np in nps])
    ylabels = [np.title.replace("ttbar", "$t\\bar{t}$").replace("tW", "$tW$") for np in nps]

    fig, ax = plt.subplots(figsize=(7, 1.0 + len(yval) * 0.38))
    fig.subplots_adjust(left=0.5, right=0.95)
    ax.fill_betweenx([-50, 500], -2, 2, color="yellow")
    ax.fill_betweenx([-50, 500], -1, 1, color="limegreen")
    ax.set_yticks(yval)
    ax.set_yticklabels(ylabels)
    ax.errorbar(
        xval,
        yval,
        xerr=[abs(xerr_lo), xerr_hi],
        fmt="ko",
        capsize=3.5,
        lw=2,
        elinewidth=2.25,
    )
    ax.set_xlim([-2.2, 2.2])
    ax.set_ylim([0.0, len(yval) + 1])
    ax.grid(color="black", alpha=0.15)

    # fmt: off
    if not args.no_text:
        for mean, iyval, minus, plus in zip(xval, yval, xerr_lo, xerr_hi):
            ax.text(mean, iyval + Y_OFFSET_TEXT_MEAN, "${}$".format(round(mean, 3)),
                    color="black", size=10, horizontalalignment="center")
            ax.text(mean + minus - 6.5 * X_OFFSET_TEXT, iyval + Y_OFFSET_TEXT, "${}$".format(round(minus, 3)),
                    color="red", size=10, horizontalalignment="center")
            ax.text(mean + plus + X_OFFSET_TEXT, iyval + Y_OFFSET_TEXT, "${}$".format(round(plus, 3)),
                    color="blue", size=10)
    # fmt: on

    fig.subplots_adjust(left=0.45)
    if nnps < 10:
        fig.subplots_adjust(bottom=(0.225 + 0.25 / nnps))
    ax.set_xlabel(r"$\left(\hat\theta - \theta_0\right) / \Delta \theta$")
    return fig, ax


def run_pulls(args: argparse.Namespace) -> None:
    """Given command line arguments generate pull plots

    Parameters
    ----------
    args : argparse.Namespace

    """
    setup_style()
    systematics, categories = get_blank_systematics(args.config)
    fit_name = PosixPath(args.workspace).stem
    fit_result = PosixPath(f"{args.workspace}/Fits/{fit_name}.txt")
    np_by_cat = {c: [] for c in categories}
    with fit_result.open("r") as f:
        lines = f.read().split("CORRELATION_MATRIX")[0].strip()
        for line in lines.split("\n")[2:-1]:
            if line.startswith("gamma"):
                continue
            elements = line.split()
            systematics[elements[0]].mean = float(elements[1])
            systematics[elements[0]].plus = float(elements[2])
            systematics[elements[0]].minus = float(elements[3])
            np_by_cat[systematics[elements[0]].category].append(systematics[elements[0]])

    if args.out_dir is None:
        outd = f"{args.workspace}/MPL"
    else:
        outd = args.out_dir
    if outd != ".":
        PosixPath(outd).mkdir(parents=True, exist_ok=True)

    for category, nps in np_by_cat.items():
        fig, ax = draw_pulls(args, nps)
        out_name = f"{outd}/pulls_{category}.pdf"
        fig.savefig(out_name)
        log.info(f"Done with {category}")



"""
a module to house some constants
"""

SELECTION_1j1b = "(reg1j1b == True) & (OS == True)"
"""
str: The pandas flavor selection string for the 1j1b region
"""


SELECTION_2j1b = "(reg2j1b == True) & (OS == True)"
"""
str: The pandas flavor selection string for the 2j1b region
"""


SELECTION_2j2b = "(reg2j2b == True) & (OS == True)"
"""
str: The pandas flavor selection string for the 2j2b region
"""


FEATURESET_1j1b_DR = sorted(
    [
        "pT_lep1",
        "pT_lep2",
    ],
    key=str.lower,
)
"""
list(str) : set of features used in the 1j1b DR BDT
"""

FEATURESET_2j1b_DR = sorted(
    [
        "pT_lep1",
        "pT_lep2",
    ],
    key=str.lower,
)
"""
list(str) : set of features used in the 2j1b DR BDT
"""


FEATURESET_2j2b_DR = sorted(
    [
        "pT_lep1",
        "pT_lep2",
    ],
    key=str.lower,
)
"""
list(str) : set of features used in the 2j2b DR BDT
"""


FEATURESET_1j1b_DS = sorted(
    [
        "pT_lep1",
        "pT_lep2",
    ],
    key=str.lower,
)
"""
list(str) : set of features used in the 1j1b DS BDT
"""


FEATURESET_2j1b_DS = sorted(
    [
        "pT_lep1",
        "pT_lep2",
    ],
    key=str.lower,
)
"""
list(str) : set of features used in the 2j1b DS BDT
"""


FEATURESET_2j2b_DS = sorted(
    [
        "pT_lep1",
        "pT_lep2",
    ],
    key=str.lower,
)
"""
list(str) : set of features used in the 2j2b DS BDT
"""


FEATURESET_1j1b = sorted(
    [
        "cent_lep1lep2",
        "deltapT_lep1_lep2",
        "mass_lep1jet1",
        "mass_lep2jet1",
        "mass_lep2jet1met",
        "mT_jet1met",
        "nsoftbjets",
        "nsoftjets",
        "pT_jetS1",
        "pTsys_jet1met",
        "pTsys_lep1lep2",
        "pTsys_lep1lep2jet1met"
    ],
    key=str.lower,
)
"""
list(str): list of features we use for classifiers in the 1j1b region
"""


FEATURESET_2j1b = sorted(
    [
        "deltaR_lep1lep2_jet1jet2met",
        "HT_jet1jet2",
        "mass_lep1jet1",
        "mass_lep1jet2",
        "mass_lep2jet1",
        "mass_lep2jet2",
        "psuedoContTagBin_jet1",
        "psuedoContTagBin_jet2",
        "pT_jet2",
        "pTsys_lep1lep2jet1jet2met",
    ],
    key=str.lower,
)
"""
list(str): list of features we use for classifiers in the 2j1b region
"""


FEATURESET_2j2b = sorted(
    [
        "deltaR_jet1_jet2",
        "mass_lep1jet1",
        "mass_lep1jet2",
        "mass_lep2jet1",
        "mass_lep2jet2",
        "pT_jet2",
        "pTsys_jet1jet2",
    ],
    key=str.lower,
)
"""
list(str): list of features we use for classifiers in the 2j2b region
"""

FEATURESET_1j1b_TMVA = sorted(
    [
        "pTsys_lep1lep2jet1met",
        "mass_lep2jet1",
        "mass_lep1jet1",
        "pTsys_lep1lep2",
        "deltaR_lep2_jet1",
        "nsoftjets",
        "deltaR_lep1_lep2",
        "deltapT_lep1_jet1",
        "mT_lep2met",
        "nsoftbjets",
        "cent_lep1lep2",
        "pTsys_lep1lep2jet1",
    ],
    key=str.lower,
)
"""
list(str): list of features we use for classifiers in the 1j1b region
"""


FEATURESET_2j1b_TMVA = sorted(
    [
        "mass_lep1jet2",
        "psuedoContTagBin_jet1",
        "mass_lep1jet1",
        "mass_lep2jet1",
        "mass_lep2jet2",
        "pTsys_lep1lep2jet1jet2met",
        "psuedoContTagBin_jet2",
        "pT_jet2",
    ],
    key=str.lower,
)
"""
list(str): list of features we use for classifiers in the 2j1b region
"""


FEATURESET_2j2b_TMVA = sorted(
    [
        "mass_lep1jet2",
        "mass_lep1jet1",
        "deltaR_lep1_jet1",
        "mass_lep2jet1",
        "pTsys_lep1lep2met",
        "pT_jet2",
        "mass_lep2jet2",
    ],
    key=str.lower,
)
"""
list(str): list of features we use for classifiers in the 2j2b region
"""


AVOID_IN_CLF = sorted(
    [
        "bdt_response",
        "eta_met",
        "eta_jetL1",
        "eta_jetS1",
        "sumet",
        "mass_jet1",
        "mass_jet2",
        "mass_jetF",
        "mass_jetL1",
        "mass_jetS1",
        "E_jetL1",
        "E_jetS1",
        "E_jet1",
        "E_jet2",
        "pT_lep3",
        "pT_jetL1",
        "nbjets",
        "njets",
    ],
    key=str.lower,
)
"""
list(str): list of features to avoid in classifiers
"""



"""
Module for handling dataframes
"""

from __future__ import annotations

# stdlib
import logging
from dataclasses import dataclass, field

# externals
import dask
import dask.dataframe as dd
import pandas as pd
import uproot

# tdub
from tdub.constants import AVOID_IN_CLF
from tdub.utils import (
    Region,
    categorize_branches,
    conservative_branches,
    get_branches,
    get_features,
    get_selection,
)


log = logging.getLogger(__name__)


class DataFramesInMemory:
    """A dataset structured with everything living on RAM

    Parameters
    ----------
    name : str
        dataset name
    ddf : :obj:`dask.dataframe.DataFrame`
        dask dataframe with all information (normal payload and
        weights)
    dropnonkin : bool
        drop columns that are not kinematic information (e.g. ``OS``
        or ``reg2j1b``)
    skeleton : bool
        initialize class with empty dataframes, ddf is ignored if ``True``.
    consolidate : bool
        call the ``consolidate`` function at the end of initialization

    Attributes
    ----------
    name : str
       dataset name
    df : :obj:`pandas.DataFrame`
       payload dataframe, for meaningful kinematic features
    weights : :obj:`pandas.DataFrame`
       dataframe to hold weight information

    Examples
    --------
    Manually constructing in memory dataframes from a dask dataframe:

    >>> from tdub.utils import quick_files
    >>> from tdub.frames import DataFramesInMemory, delayed_dataframe
    >>> ttbar_files = quick_files("/path/to/data")["ttbar"]
    >>> branches = ["pT_lep1", "met", "mass_lep1jet1"]
    >>> ddf = delayed_dataframe(ttbar_files, branches=branches)
    >>> dfim = DataFramesInMemory(name="ttbar", ddf=ddf)

    Having the legwork done by other module features (see
    :py:func:`specific_dataframe`):

    >>> from tdub.frames import specific_dataframe
    >>> dfim = specific_dataframe(ttbar_files, "2j2b", to_ram=True)

    """

    def __init__(
        self,
        *,
        name: str = "",
        ddf: Optional[dask.dataframe.DataFrame] = None,
        skeleton: bool = False,
        dropnonkin: bool = False,
        consolidate: bool = False,
    ) -> None:
        if skeleton:
            self.name = name
            self._df = pd.DataFrame()
            self._weights = pd.DataFrame()
        else:
            if ddf is None:
                raise ValueError("if skeleton=False, a dask dataframe is required")
            self.name = name
            all_columns = list(ddf.columns)
            categorized = categorize_branches(all_columns)
            nonweights = categorized["kinematics"]
            if not dropnonkin:
                nonweights += categorized["meta"]
            self._df = ddf[sorted(nonweights)].compute()
            self._weights = ddf[categorized["weights"]].compute()
            if consolidate:
                self.consolidate()

    @property
    def df(self):
        return self._df

    @property
    def weights(self):
        return self._weights

    def __repr__(self):
        if self.weights is not None:
            return "DataFramesInMemory(name={}, df_shape={}, weights_shape={})".format(
                self.name, self.df.shape, self.weights.shape
            )
        else:
            return "DataFramesInMemory(name={}, df_shape={}, weights=None)".format(
                self.name, self.df.shape
            )

    def consolidate(self) -> None:
        """consolidate the data into a single dataframe

        this will remove the weights from the separate dataframe and
        move the data to the ``df`` attribute; the ``weights attribute
        will be reassigned to ``None``.
        """
        weight_cols = list(self.weights.columns)
        for weight_col in weight_cols:
            log.info(f"Moving {weight_col} to the main dataframe")
            self._df[weight_col] = self._weights.pop(weight_col)
        self._weights = None


@dataclass
class SelectedDataFrame:
    """DataFrame constructed from a selection string

    Attributes
    ----------
    name : str
       shorthand name of the selection
    selection : str
       the selection string (in :py:func:`pandas.eval` form)
    df : :obj:`dask.dataframe.DataFrame`
       the dask DataFrame

    Notes
    -----
    This class is not designed for instance creation on the user end,
    we have factory functions for creating instances

    See Also
    --------
    specific_dataframe
    selected_dataframes

    """

    name: str
    selection: str
    df: dd.DataFrame = field(repr=False, compare=False)

    def to_ram(self, **dfim_opts) -> DataFramesInMemory:
        """create a dataset that lives in memory

        Parameters
        ----------
        dfim_opts
           options passed to the :py:obj:`DataFramesInMemory`
           initializer

        Returns
        -------
        :obj:`DataFramesInMemory`
           the dataset now living in memory

        Examples
        --------

        >>> from tdub.utils import quick_files
        >>> from tdub.frames import specific_dataframe
        >>> files = quick_files("/path/to/data")["ttbar"]
        >>> sdf = specific_dataframe(files, "2j2b", name="ttbar_2j2b")
        >>> dfim = sdf.to_ram(dropnonkin=False)

        """
        return DataFramesInMemory(name=self.name, ddf=self.df, **dfim_opts)


def raw_dataframe(
    files: Union[str, List[str]],
    tree: str = "WtLoop_nominal",
    weight_name: str = "weight_nominal",
    branches: Optional[List[str]] = None,
    entrysteps: Optional[Any] = None,
) -> pandas.DataFrame:
    """Construct a raw pandas flavored Dataframe with help from uproot

    We call this dataframe "raw" because it hasn't been parsed by any
    other tdub.frames functionality (no selection performed, kinematic
    and weight branches won't be separated, etc.) -- just a pure raw
    dataframe from some ROOT files.

    Parameters
    ----------
    files : list(str) or str
       a single ROOT file or list of ROOT files
    tree : str
       the tree name to turn into a dataframe
    weight_name: str
       weight branch (we make sure to grab it if you give something
       other than ``None`` to ``branches``).
    branches : list(str), optional
       a list of branches to include as columns in the dataframe,
       default is ``None``, includes all branches.
    entrysteps : Any, optional
       see the ``entrysteps`` keyword for ``uproot.iterate``

    Returns
    -------
    :obj:`pandas.DataFrame`
       the pandas flavored DataFrame with all requested branches

    Examples
    --------

    >>> from tdub.utils import quick_files
    >>> from tdub.frames import raw_dataframe
    >>> files = quick_files("/path/to/files")["ttbar"]
    >>> df = raw_dataframe(files)

    """
    bs = branches
    if branches is not None:
        bs = sorted(set(branches) | set([weight_name]), key=str.lower)
    return pd.concat(
        [d for d in uproot.pandas.iterate(files, tree, branches=bs, entrysteps=entrysteps)]
    )


def conservative_dataframe(
    files: Union[str, List[str]],
    tree: str = "WtLoop_nominal",
    weight_name: str = "weight_nominal",
    entrysteps: Optional[Any] = None,
) -> pandas.DataFrame:
    """Construct a raw pandas flavored dataframe with conservative branches

    This function does some hand-holding and grabs a conservative set
    of branches from the input file(s). The branches that will be
    columns in the dataframe are determined by
    :py:func:`tdub.utils.conservative_branches`.

    Parameters
    ----------
    files : list(str) or str
       a single ROOT file or list of ROOT files
    tree : str
       the tree name to turn into a dataframe
    weight_name: str
       weight branch (we make sure to grab it)
    entrysteps : Any, optional
       see the ``entrysteps`` keyword for ``uproot.iterate``

    Returns
    -------
    :obj:`pandas.DataFrame`
       the pandas flavored DataFrame with all requested branches

    Examples
    --------

    >>> from tdub.utils import quick_files
    >>> from tdub.frames import conservative_dataframe
    >>> files = quick_files("/path/to/files")["ttbar"]
    >>> df = conservative_dataframe(files)

    """
    if isinstance(files, str):
        bs = conservative_branches(files, tree)
    else:
        bs = conservative_branches(files[0], tree)
    bs = list(set(bs) | set([weight_name]))
    return raw_dataframe(
        files, tree=tree, weight_name=weight_name, entrysteps=entrysteps, branches=bs
    )


def delayed_dataframe(
    files: Union[str, List[str]],
    tree: str = "WtLoop_nominal",
    weight_name: str = "weight_nominal",
    branches: Optional[List[str]] = None,
    repartition_kw: Optional[Dict[str, Any]] = None,
    experimental: bool = False,
) -> dask.dataframe.DataFrame:
    """Construct a dask flavored DataFrame with help from uproot

    Parameters
    ----------
    files : list(str) or str
       a single ROOT file or list of ROOT files
    tree : str
       the tree name to turn into a dataframe
    weight_name: str
       weight branch (we make sure to grab it if you give something
       other than ``None`` to ``branches``).
    branches : list(str), optional
       a list of branches to include as columns in the dataframe,
       default is ``None``, includes all branches.
    repartition_kw : dict(str, Any), optional
       arguments to pass to :py:func:`dask.dataframe.DataFrame.repartition`
    experimental: bool
       if ``True`` try letting uproot create the Dask DataFrame instead of
       using ``dask.delayed`` on pandas DataFrames grabbed via uproot.

    Returns
    -------
    :obj:`dask.dataframe.DataFrame`
       the dask flavored DataFrame with all requested branches

    Examples
    --------

    >>> from tdub.utils import quick_files
    >>> from tdub.frames import delayed_dataframe
    >>> files = quick_files("/path/to/files")["tW_DR"]
    >>> ddf = delayed_dataframe(files, branches=["branch_a", "branch_b"])

    """
    bs = branches
    if branches is not None:
        bs = sorted(set(branches) | set([weight_name]), key=str.lower)

    # fmt: off
    if experimental:
        print("using experimental dataframe creation via uproot.daskframe")
        import cachetools
        cache = cachetools.LRUCache(1)
        ddf = uproot.daskframe(files, tree, bs, namedecode="utf-8", basketcache=cache)
    else:
        @dask.delayed
        def get_frame(f, tn):
            t = uproot.open(f)[tn]
            return t.pandas.df(branches=bs)
        if isinstance(files, str):
            dfs = [get_frame(files, tree)]
        else:
            dfs = [get_frame(f, tree) for f in files]
        ddf = dd.from_delayed(dfs)
    # fmt: on

    if repartition_kw is not None:
        log.info(f"repartition with {repartition_kw}")
        ddf = ddf.repartition(**repartition_kw)
    return ddf


def selected_dataframes(
    files: Union[str, List[str]],
    selections: Dict[str, str],
    tree: str = "WtLoop_nominal",
    weight_name: str = "weight_nominal",
    branches: Optional[List[str]] = None,
    delayed_dataframe_kw: Optional[Dict[str, Any]] = None,
) -> Dict[str, SelectedDataFrame]:
    """Construct a set of dataframes based on a list of selection queries

    Parameters
    ----------
    files : list(str) or str
       a single ROOT file or list of ROOT files
    selections : dict(str,str)
       the list of selections to apply on the dataframe in the form
       ``(name, query)``.
    tree : str
       the tree name to turn into a dataframe
    weight_name: str
       weight branch
    branches : list(str), optional
       a list of branches to include as columns in the dataframe,
       default is ``None`` (all branches)
    delayed_dataframe_kw : dict(str, Any), optional
       set of arguments to pass to :py:func:`delayed_dataframe`

    Returns
    -------
    dict(str, :obj:`SelectedDataFrame`)
       dictionary containing queried dataframes.

    Examples
    --------

    >>> from tdub.utils import quick_files
    >>> from tdub.frames import selected_dataframes
    >>> files = quick_files("/path/to/files")["tW_DS"]
    >>> selections = {"r2j2b": "(reg2j2b == True) & (OS == True)",
    ...               "r2j1b": "(reg2j1b == True) & (OS == True)"}
    >>> frames = selected_dataframes(files, selections=selections)

    """
    if delayed_dataframe_kw is None:
        df = delayed_dataframe(files, tree, weight_name, branches)
    else:
        df = delayed_dataframe(files, tree, weight_name, branches, **delayed_dataframe_kw)
    return {
        sel_name: SelectedDataFrame(sel_name, sel_query, df.query(sel_query))
        for sel_name, sel_query in selections.items()
    }


def specific_dataframe(
    files: Union[str, List[str]],
    region: Union[Region, str],
    name: str = "nameless",
    tree: str = "WtLoop_nominal",
    weight_name: str = "weight_nominal",
    extra_branches: Optional[List[str]] = None,
    bypass_dask: bool = False,
    to_ram: bool = False,
    **kwargs,
) -> Union[SelectedDataFrame, DataFramesInMemory]:
    """Construct a dataframe based on specific predefined region selection

    Parameters
    ----------
    files : list(str) or str
       a single ROOT file or list of ROOT files
    region : tdub.utils.Region or str
       which predefined tW region to select
    name : str
       give your selection a name
    tree : str
       the tree name to turn into a dataframe
    weight_name: str
       weight branch
    extra_branches : list(str), optional
       a list of additional branches to save (the standard branches
       associated as features for the region you selected will be
       included by default).
    bypass_dask: bool
       bypass going through a delayed dataframe, just use pandas; this
       will cause the function to return a :obj:`DataFramesInMemory`
       instance (``to_ram`` will be ignored)
    to_ram : bool
       automatically send dataset to memory via
       :py:func:`SelectedDataFrame.to_ram`

    Keyword Args
    ------------
    consolidate : bool
       consolidate weight columns into the main DataFrame (usually
       they are seperate, see :py:obj:`DataFramesInMemory`)
    dropnonkin : bool
       drop non-kinematic columns (metadata like columns, like ``OS``,
       ``reg2j1b``, etc.).

    Returns
    -------
    :obj:`SelectedDataFrame` or :obj:`DataFramesInMemory`
       if ``to_ram`` is ``False``, we return the dask-backed ``SelectedDataFrame``,
       if ``to_ram`` is ``True``, we return the pandas-backed ``DataFramesInMemory``.
       if ``pybass_dask`` is ``True``, we return the pandas-backed ``DataFramesInMemory``.

    Examples
    --------

    >>> from tdub.utils import quick_files
    >>> from tdub.frames import specific_dataframe
    >>> files = quick_files("/path/to/files")["ttbar"]
    >>> frame_2j1b = specific_dataframe(files, Region.r2j1b, extra_branches=["pT_lep1"])
    >>> frame_2j2b = specific_dataframe(files, "2j2b", extra_branches=["met"])

    Here we bypass dask for converting the ROOT files to a dataframe
    (this is faster for smaller dataframes, for large dataframes
    (i.e. ttbar samples) we should stick to dask.

    >>> frame_2j2b = specific_dataframe(files, "2j2b", bypass_dask=True)

    Ask that the weight branches/columns are stored in the main
    ``df`` DataFrame (not in the ``weights`` DataFrame).

    >>> from tdub.frames import specific_dataframe
    >>> dfim = specific_dataframe(ttbar_files, "2j2b", to_ram=True, consolidate=True)
    >>> dfim.weights is None:
    True

    Or perform the consolidation independent of initialization:

    >>> from tdub.frames import specific_dataframe
    >>> dfim = specific_dataframe(ttbar_files, "2j2b", to_ram=True)
    >>> dfim.consolidate()

    """
    if isinstance(region, str):
        reg = Region.from_str(region)
    elif isinstance(region, Region):
        reg = region
    else:
        raise TypeError("region argument must be tdub.utils.Region or str")
    region_branches = get_features(reg)
    if extra_branches is None:
        extra_branches = []
    if reg == Region.r1j1b:
        branches = sorted(
            set(region_branches) | set(extra_branches) | {"reg1j1b", "OS"}, key=str.lower
        )
    elif reg == Region.r2j1b:
        branches = sorted(
            set(region_branches) | set(extra_branches) | {"reg2j1b", "OS"}, key=str.lower
        )
    elif reg == Region.r2j2b:
        branches = sorted(
            set(region_branches) | set(extra_branches) | {"reg2j2b", "OS"}, key=str.lower
        )

    q = get_selection(reg)

    if bypass_dask:
        raw_df = raw_dataframe(files, tree, weight_name, branches=branches)
        weight_df = pd.DataFrame(raw_df.pop(weight_name))
        sel = raw_df.eval(q)
        dfim = DataFramesInMemory(skeleton=True)
        categorized = categorize_branches(raw_df)
        use_branches, meta_branches = categorized["kinematics"], categorized["meta"]
        if not kwargs.pop("dropnonkin", False):
            use_branches += meta_branches
        dfim._df = raw_df[sel][sorted(use_branches, key=str.lower)]
        dfim._weights = weight_df[sel]
        if kwargs.pop("consolidate", False):
            dfim.consolidate()
        return dfim

    sdf = SelectedDataFrame(
        name, q, delayed_dataframe(files, tree, weight_name, branches).query(q)
    )
    if to_ram:
        return sdf.to_ram(**kwargs)
    return sdf


def satisfying_selection(*dfs: pandas.DataFrame, selection: str) -> List[pandas.DataFrame]:
    """get subsets of dataframes that satisfy a selection

    Parameters
    ----------
    *dfs : sequence of :py:obj:`pandas.DataFrame`
       the dataframes to apply the selection to
    selection : str
       the selection string (in :py:func:`pandas.eval` form)

    Returns
    -------
    list(pandas.DataFrame)
       the dataframes satisfying the selection string

    Examples
    --------

    >>> from tdub.utils import quick_files
    >>> from tdub.frames import specific_dataframe, satisfying_selection
    >>> qf = quick_files("/path/to/files")
    >>> dfim_tW_DR = specific_dataframe(qf["tW_DR"], to_ram=True)
    >>> dfim_ttbar = specific_dataframe(qf["ttbar"], to_ram=True)
    >>> low_bdt = "(bdt_response < 0.4)"
    >>> tW_DR_selected, ttbar_selected = satisfying_selection(
    ...     dfim_tW_DR.df, dfim_ttbar.df, selection=low_bdt
    ... )

    """
    return [df.query(selection) for df in dfs]


def iterative_selection(
    files: Union[str, List[str]],
    selection: str,
    tree: str = "WtLoop_nominal",
    weight_name: str = "weight_nominal",
    branches: Optional[List[str]] = None,
    concat: bool = False,
    keep_category: Optional[str] = None,
    ignore_avoid: bool = False,
    **iterate_opts,
) -> pandas.DataFrame:
    """build a selected dataframe via uproot's iterate

    if we want to build a memory-hungry dataframe and apply a
    selection this helps us avoid crashing due to using all of our
    RAM.

    ``iterate_opts`` are fed to :py:func:`uproot.pandas.iterate`

    this dataframe construction function is useful when we want to
    grab all of the branches in a large dataset that won't fit in
    memory before the selection.

    Parameters
    ----------
    files : list(str) or str
       a single ROOT file or list of ROOT files
    selection : str
       the selection string (in :py:func:`pandas.eval` form)
    tree : str
       the tree name to turn into a dataframe
    weight_name: str
       weight branch to preserve
    branches : list(str), optional
       a list of branches to include as columns in the dataframe,
       default is ``None`` (all branches)
    concat : bool
       concatenate the resulting selected dataframes to return
       a single dataframe
    keep_category : str, optional
       if not ``None``, the selected dataframe(s) will only include
       columns which are part of the given category (see
       :py:func:`tdub.utils.categorize_branches`). The weight branch
       is always kept.
    ignore_avoid : bool
       ignore branches defined by :py:data:`tdub.utils.AVOID_IN_CLF`

    Returns
    -------
    list(pandas.DataFrame) or pandas.DataFrame
       the final selected dataframe(s) from the files

    Examples
    --------

    Creating a ``ttbar_dfs`` list of dataframes and a single ``tW_df``
    dataframe:

    >>> from tdub.frames import iterative_selection
    >>> from tdub.utils import quick_files
    >>> from tdub.utils import get_selection
    >>> qf = quick_files("/path/to/files")
    >>> ttbar_dfs = iterative_selection(qf["ttbar"], get_selection("2j2b"), entrysteps="1 GB")
    >>> tW_df = iterative_selection(qf["tW_DR"], get_selection("2j2b"), concat=True)

    Keep only the kinematic branches after selection and ignore avoided columns:

    >>> tW_df = iterative_selection(qf["tW_DR"], get_selection("2j2b"), concat=True,
    ...                             keep_category="kinematics", ignore_avoid=True)

    """
    if keep_category is not None:
        if isinstance(files, list):
            branch_cats = categorize_branches(files[0], tree=tree)
        else:
            branch_cats = categorize_branches(files, tree=tree)
        keep_branches = branch_cats.get(keep_category)
        if weight_name not in keep_branches:
            keep_branches.append(weight_name)

    bs = branches
    if branches is not None:
        bs = sorted(set(branches) | set([weight_name]), key=str.lower)

    if ignore_avoid:
        if bs is None:
            if isinstance(files, list):
                filebs = get_branches(files[0], tree=tree)
            else:
                filebs = get_branches(files, tree=tree)
            bs = sorted(set(filebs) - set(AVOID_IN_CLF), key=str.lower)
        else:
            bs = sorted(set(bs) - set(AVOID_IN_CLF), key=str.lower)
        if keep_category is not None:
            keep_branches = sorted(set(keep_branches) & set(bs), key=str.lower)

    dfs = []
    itr = uproot.pandas.iterate(files, tree, branches=bs, **iterate_opts)
    for i, df in enumerate(itr):
        idf = df.query(selection)
        if keep_category is not None:
            idf = idf[keep_branches]
        dfs.append(idf)
        log.debug(f"finished iteration {i}")
    if concat:
        return pd.concat(dfs)
    return dfs


def drop_cols(df: pandas.DataFrame, *cols: str) -> None:
    """drop some columns from a dataframe

    this is a convenient function because it just ignores branches
    that don't exist in the dataframe that are present in ``cols``.

    we augment :py:class:`pandas.DataFrame` with this function

    Parameters
    ----------
    df : :py:obj:`pandas.DataFrame`
       the df which we want to slim
    *cols : sequence of strings
       the columns to remove

    Examples
    --------

    >>> import pandas as pd
    >>> from tdub.utils import drop_cols
    >>> df = pd.read_parquet("some_file.parquet")
    >>> "E_jet1" in df.columns:
    True
    >>> "mass_jet1" in df.columns:
    True
    >>> "mass_jet2" in df.columns:
    True
    >>> drop_cols(df, "E_jet1", "mass_jet1")
    >>> "E_jet1" in df.columns:
    False
    >>> "mass_jet1" in df.columns:
    False
    >>> df.drop_cols("mass_jet2") # use augmented df class
    >>> "mass_jet2" in df.columns:
    False

    """
    in_dataframe = set(df.columns)
    in_cols = set(cols)
    in_both = list(in_dataframe & in_cols)
    df.drop(columns=in_both, inplace=True)


pd.DataFrame.drop_cols = drop_cols


def drop_avoid(df: pandas.DataFrame) -> None:
    """drop columns that we avoid in classifiers

    this uses :py:func:`tdub.frames.drop_cols` with a predefined set
    of columns (:py:data:`tdub.utils.AVOID_IN_CLF`).

    we augment :py:class:`pandas.DataFrame` with this function

    Parameters
    ----------
    df : :py:obj:`pandas.DataFrame`
       the df which we want to slim

    Examples
    --------

    >>> from tdub.utils import drop_avoid
    >>> import pandas as pd
    >>> df = pd.read_parquet("some_file.parquet")
    >>> "E_jetL1" in df.columns:
    True
    >>> drop_avoid(df)
    >>> "E_jetL1" in df.columns:
    False

    """
    drop_cols(df, *AVOID_IN_CLF)


pd.DataFrame.drop_avoid = drop_avoid


def drop_jet2(df: pandas.DataFrame) -> None:
    """drop all columns with jet2 properties

    in the 1j1b region we obviously don't have a second jet; so this
    lets us get rid of all columns dependent on jet2 kinematic
    properties.

    we augment :py:class:`pandas.DataFrame` with this function

    Parameters
    ----------
    df : :py:obj:`pandas.DataFrame`
       the df which we want to slim

    Examples
    --------

    >>> from tdub.utils import drop_jet2
    >>> import pandas as pd
    >>> df = pd.read_parquet("some_file.parquet")
    >>> "pTsys_lep1lep2jet1jet2met" in df.columns:
    True
    >>> drop_jet2(df)
    >>> "pTsys_lep1lep2jet1jet2met" in df.columns:
    False

    """
    j2cols = [col for col in df.columns if "jet2" in col]
    drop_cols(df, *j2cols)


pd.DataFrame.drop_jet2 = drop_jet2



"""
Module for training BDTs
"""

from __future__ import annotations

# stdlib
import json
import logging
import os
from pathlib import PosixPath
from pprint import pformat

# externals
import joblib
import lightgbm as lgbm
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import pygram11
from scipy import interp
from sklearn.model_selection import KFold, train_test_split
from sklearn.metrics import auc, roc_auc_score, roc_curve

# tdub
from tdub.frames import specific_dataframe
from tdub.utils import Region, bin_centers, quick_files, ks_twosample_binned


log = logging.getLogger(__name__)


def prepare_from_root(
    sig_files: List[str],
    bkg_files: List[str],
    region: Union[Region, str],
    weight_scale: float = 1.0e3,
    scale_sum_weights: bool = True,
) -> Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]:
    """Prepare the data for training in a region with signal and
    background ROOT files

    Parameters
    ----------
    sig_files : list(str)
       list of signal ROOT files
    bkg_files : list(str)
       list of background ROOT files
    region : Region or str
       the region where we're going to perform the training
    weight_scale : float
       value to scale all weights by
    scale_sum_weights : bool
       scale sum of weights of signal to be sum of weights of background

    Returns
    -------
    df : :obj:`pandas.DataFrame`
       the feature matrix
    labels : :obj:`numpy.ndarray`
       the event labels (``0`` for background; ``1`` for signal)
    weights : :obj:`numpy.ndarray`
       the event weights

    Examples
    --------

    >>> from tdub.utils import quick_files
    >>> from tdub.train import prepare_from_root
    >>> qfiles = quick_files("/path/to/data")
    >>> df, labels, weights = prepare_from_root(qfiles["tW_DR"], qfiles["ttbar"], "2j2b")

    """
    log.info("preparing training data")
    log.info("signal files:")
    for f in sig_files:
        log.info(f"  - {f}")
    log.info("background files:")
    for f in bkg_files:
        log.info(f"  - {f}")

    ## signal pretty much always be tW, no need for dask
    sig_dfim = specific_dataframe(
        sig_files, region, "train_sig", bypass_dask=True, dropnonkin=True
    )
    ## bkg is pretty much always ttbar, so lets use daks to be careful
    bkg_dfim = specific_dataframe(
        bkg_files, region, "train_bkg", to_ram=True, dropnonkin=True
    )

    sorted_cols = sorted(sig_dfim.df.columns.to_list(), key=str.lower)
    sig_dfim._df = sig_dfim._df[sorted_cols]
    bkg_dfim._df = bkg_dfim._df[sorted_cols]

    cols = sig_dfim.df.columns.to_list()
    assert cols == bkg_dfim.df.columns.to_list(), "sig/bkg columns are different. bad."
    log.info("features used:")
    for c in cols:
        log.info(f"  - {c}")

    w_sig = sig_dfim.weights.weight_nominal.to_numpy()
    w_bkg = bkg_dfim.weights.weight_nominal.to_numpy()
    w_sig[w_sig < 0] = 0.0
    w_bkg[w_bkg < 0] = 0.0
    w_sig *= weight_scale
    w_bkg *= weight_scale
    if scale_sum_weights:
        w_sig *= w_bkg.sum() / w_sig.sum()

    df = pd.concat([sig_dfim.df, bkg_dfim.df])
    y = np.concatenate([np.ones_like(w_sig), np.zeros_like(w_bkg)])
    w = np.concatenate([w_sig, w_bkg])

    return df, y, w


def folded_training(
    df: pandas.DataFrame,
    labels: numpy.ndarray,
    weights: numpy.ndarray,
    params: Dict[str, Any],
    fit_kw: Dict[str, Any],
    output_dir: Union[str, os.PathLike],
    region: str,
    kfold_kw: Dict[str, Any] = None,
) -> float:
    """Train a :obj:`lightgbm.LGBMClassifier` model using :math:`k`-fold
    cross validation using the given input data and parameters.  The
    models resulting from the training (and other important training
    information) are saved to ``output_dir``. The entries in the
    ``kfold_kw`` argument are forwarded to the
    :obj:`sklearn.model_selection.KFold` class for data
    preprocessing. The default arguments that we use are:

    - ``n_splits``: 3
    - ``shuffle``: ``True``
    - ``random_state``: 414

    Parameters
    ----------
    df : :obj:`pandas.DataFrame`
       the feature matrix in dataframe format
    labels : :obj:`numpy.ndarray`
       the event labels (``1`` for signal; ``0`` for background)
    weights : :obj:`numpy.ndarray`
       the event weights
    params : dict(str, Any)
       dictionary of :obj:`lightgbm.LGBMClassifier` parameters
    fit_kw : dict(str, Any)
       dictionary of arguments forwarded to :py:func:`lightgbm.LGBMClassifier.fit`.
    output_dir : str or os.PathLike
       directory to save results of training
    region : str
        string representing the region
    kfold_kw : optional dict(str, Any)
       arguments fed to :obj:`sklearn.model_selection.KFold`

    Returns
    -------
    neg_roc_score : float
       -1 times the mean area under the ROC curve (AUC)

    Examples
    --------

    >>> from tdub.utils import quick_files
    >>> from tdub.train import prepare_from_root
    >>> from tdub.train import folded_training
    >>> qfiles = quick_files("/path/to/data")
    >>> df, labels, weights = prepare_from_root(qfiles["tW_DR"], qfiles["ttbar"], "2j2b")
    >>> params = dict(
    ...     boosting_type="gbdt",
    ...     num_leaves=42,
    ...     learning_rate=0.05
    ...     reg_alpha=0.2,
    ...     reg_lambda=0.8,
    ...     max_depth=5,
    ... )
    >>> folded_training(
    ...     df,
    ...     labels,
    ...     weights,
    ...     params,
    ...     {"verbose": 20},
    ...     "/path/to/train/output",
    ...     "2j2b",
    ...     kfold_kw={"n_splits": 5, "shuffle": True, "random_state": 17}
    ... )

    """
    starting_dir = os.getcwd()
    output_path = PosixPath(output_dir)
    output_path.mkdir(exist_ok=True, parents=True)
    os.chdir(output_path)

    fig_proba_hists, ax_proba_hists = plt.subplots()
    fig_pred_hists, ax_pred_hists = plt.subplots()
    fig_rocs, ax_rocs = plt.subplots()

    tprs = []
    aucs = []
    importances = np.zeros((len(df.columns)))
    mean_fpr = np.linspace(0, 1, 100)
    folder = KFold(**kfold_kw)
    fold_number = 0
    nfits = 0
    for train_idx, test_idx in folder.split(df):
        X_train, X_test = df.iloc[train_idx], df.iloc[test_idx]
        y_train, y_test = labels[train_idx], labels[test_idx]
        w_train, w_test = weights[train_idx], weights[test_idx]
        validation_data = [(X_test, y_test)]
        validation_w = w_test

        n_sig = y_train[y_train == 1].shape[0]
        n_bkg = y_train[y_train == 0].shape[0]
        scale_pos_weight = n_bkg / n_sig
        log.info(f"n_bkg / n_sig = {n_bkg} / {n_sig} = {scale_pos_weight}")

        params["scale_pos_weight"] = scale_pos_weight
        model = lgbm.LGBMClassifier(**params)
        fitted_model = model.fit(
            X_train,
            y_train,
            eval_set=validation_data,
            eval_metric="auc",
            eval_sample_weight=[validation_w],
            **fit_kw,
        )

        joblib.dump(
            fitted_model, f"model_fold{fold_number}.joblib.gz", compress=("gzip", 3)
        )

        nfits += 1
        importances += fitted_model.feature_importances_

        fold_fig_proba, fold_ax_proba = plt.subplots()
        fold_fig_pred, fold_ax_pred = plt.subplots()

        test_proba = fitted_model.predict_proba(X_test)[:, 1]
        train_proba = fitted_model.predict_proba(X_train)[:, 1]
        test_pred = fitted_model.predict(X_test, raw_score=True)
        train_pred = fitted_model.predict(X_train, raw_score=True)

        proba_sig_test = test_proba[y_test == 1]
        proba_bkg_test = test_proba[y_test == 0]
        proba_sig_train = train_proba[y_train == 1]
        proba_bkg_train = train_proba[y_train == 0]
        pred_sig_test = test_pred[y_test == 1]
        pred_bkg_test = test_pred[y_test == 0]
        pred_sig_train = train_pred[y_train == 1]
        pred_bkg_train = train_pred[y_train == 0]
        w_sig_test = w_test[y_test == 1]
        w_bkg_test = w_test[y_test == 0]
        w_sig_train = w_train[y_train == 1]
        w_bkg_train = w_train[y_train == 0]

        proba_bins = np.linspace(0, 1, 41)
        proba_bc = bin_centers(proba_bins)
        pred_bins = np.linspace(0, 1, 41)
        pred_bc = bin_centers(pred_bins)
        # proba_bins = np.linspace(proba_bkg_test.min(), proba_sig_test.max(), 41)
        # proba_bc = bin_centers(proba_bins)
        # pred_bins = np.linspace(pred_bkg_test.min(), pred_sig_test.max(), 41)
        # pred_bc = bin_centers(pred_bins)

        ### Axis with all folds (proba histograms)
        ax_proba_hists.hist(
            proba_sig_test,
            bins=proba_bins,
            label=f"F{fold_number} Sig. (test)",
            density=True,
            histtype="step",
            weights=w_sig_test,
        )
        ax_proba_hists.hist(
            proba_bkg_test,
            bins=proba_bins,
            label=f"F{fold_number} Bkg. (test)",
            density=True,
            histtype="step",
            weights=w_bkg_test,
        )
        ax_proba_hists.hist(
            proba_sig_train,
            bins=proba_bins,
            label=f"F{fold_number} Sig. (train)",
            density=True,
            histtype="step",
            weights=w_sig_train,
        )
        ax_proba_hists.hist(
            proba_bkg_train,
            bins=proba_bins,
            label=f"F{fold_number} Bkg. (train)",
            density=True,
            histtype="step",
            weights=w_bkg_train,
        )

        ### Axis specific to the fold (proba histograms)
        fold_ax_proba.hist(
            proba_sig_train,
            bins=proba_bins,
            label=f"F{fold_number} Sig. (train)",
            weights=w_sig_train,
            density=True,
            histtype="stepfilled",
            color="C0",
            edgecolor="C0",
            alpha=0.5,
            linewidth=1,
        )
        fold_ax_proba.hist(
            proba_bkg_train,
            bins=proba_bins,
            label=f"F{fold_number} Bkg. (train)",
            weights=w_bkg_train,
            density=True,
            histtype="step",
            hatch="//",
            edgecolor="C3",
            linewidth=1,
        )
        train_h_sig = pygram11.histogram(
            proba_sig_test, bins=proba_bins, weights=w_sig_test, flow=False, density=True
        )
        train_h_bkg = pygram11.histogram(
            proba_bkg_test, bins=proba_bins, weights=w_bkg_test, flow=False, density=True
        )
        fold_ax_proba.errorbar(
            proba_bc,
            train_h_sig[0],
            yerr=train_h_sig[1],
            color="C0",
            fmt="o",
            label=f"F{fold_number} Sig. (test)",
            markersize=4,
        )
        fold_ax_proba.errorbar(
            proba_bc,
            train_h_bkg[0],
            yerr=train_h_bkg[1],
            color="C3",
            fmt="o",
            label=f"F{fold_number} Bkg. (test)",
            markersize=4,
        )
        fold_ax_proba.set_ylim([0, 1.5 * fold_ax_proba.get_ylim()[1]])

        ### Axis with all
        ax_pred_hists.hist(
            pred_sig_test,
            bins=pred_bins,
            label=f"F{fold_number} Sig. (test)",
            density=True,
            histtype="step",
            weights=w_sig_test,
        )
        ax_pred_hists.hist(
            pred_bkg_test,
            bins=pred_bins,
            label=f"F{fold_number} Bkg. (test)",
            density=True,
            histtype="step",
            weights=w_bkg_test,
        )
        ax_pred_hists.hist(
            pred_sig_train,
            bins=pred_bins,
            label=f"F{fold_number} Sig. (train)",
            density=True,
            histtype="step",
            weights=w_sig_train,
        )
        ax_pred_hists.hist(
            pred_bkg_train,
            bins=pred_bins,
            label=f"F{fold_number} Bkg. (train)",
            density=True,
            histtype="step",
            weights=w_bkg_train,
        )

        fold_ax_pred.hist(
            pred_sig_test,
            bins=pred_bins,
            label=f"F{fold_number} Sig. (test)",
            density=True,
            histtype="step",
            weights=w_sig_test,
        )
        fold_ax_pred.hist(
            pred_bkg_test,
            bins=pred_bins,
            label=f"F{fold_number} Bkg. (test)",
            density=True,
            histtype="step",
            weights=w_bkg_test,
        )
        fold_ax_pred.hist(
            pred_sig_train,
            bins=pred_bins,
            label=f"F{fold_number} Sig. (train)",
            density=True,
            histtype="step",
            weights=w_sig_train,
        )
        fold_ax_pred.hist(
            pred_bkg_train,
            bins=pred_bins,
            label=f"F{fold_number} Bkg. (train)",
            density=True,
            histtype="step",
            weights=w_bkg_train,
        )
        fold_ax_pred.set_ylim([0, 1.5 * fold_ax_pred.get_ylim()[1]])

        fpr, tpr, thresholds = roc_curve(y_test, test_proba, sample_weight=w_test)
        tprs.append(interp(mean_fpr, fpr, tpr))
        tprs[-1][0] = 0.0
        roc_auc = auc(fpr, tpr)
        aucs.append(roc_auc)
        ax_rocs.plot(
            fpr, tpr, lw=1, alpha=0.45, label=f"fold {fold_number}, AUC = {roc_auc:0.3}"
        )

        fold_ax_proba.legend(ncol=2, loc="upper center")
        fold_ax_pred.legend(ncol=2, loc="upper center")
        fold_fig_proba.savefig(f"fold{fold_number}_histograms_proba.pdf")
        fold_fig_pred.savefig(f"fold{fold_number}_histograms_pred.pdf")

        plt.close(fold_fig_proba)
        plt.close(fold_fig_pred)

        fold_number += 1

    relative_importances = importances / nfits
    relative_importances = relative_importances / relative_importances.sum()

    mean_tpr = np.mean(tprs, axis=0)
    mean_tpr[-1] = 1.0
    mean_auc = auc(mean_fpr, mean_tpr)
    std_auc = np.std(aucs)
    ax_rocs.plot(
        mean_fpr,
        mean_tpr,
        color="b",
        label=f"AUC = {mean_auc:0.2} $\\pm$ {std_auc:0.2}",
        lw=2,
        alpha=0.8,
    )

    ax_proba_hists.legend(ncol=3, loc="upper center", fontsize="small")
    ax_proba_hists.set_ylim([0, 1.5 * ax_proba_hists.get_ylim()[1]])
    fig_proba_hists.savefig("histograms_proba.pdf")

    ax_pred_hists.legend(ncol=3, loc="upper center", fontsize="small")
    ax_pred_hists.set_ylim([0, 1.5 * ax_pred_hists.get_ylim()[1]])
    fig_pred_hists.savefig("histograms_pred.pdf")

    ax_rocs.legend(ncol=2, loc="lower right")
    fig_rocs.savefig("roc.pdf")

    summary = {}
    summary["region"] = region
    summary["features"] = [str(c) for c in df.columns]
    summary["importances"] = list(relative_importances)
    summary["kfold"] = kfold_kw
    summary["roc"] = {
        "auc": mean_auc,
        "std": std_auc,
        "mean_fpr": list(mean_fpr),
        "mean_tpr": list(mean_tpr),
    }

    with open("summary.json", "w") as f:
        json.dump(summary, f, indent=4)

    os.chdir(starting_dir)
    neg_roc_score = -1.0 * np.mean(aucs)
    return neg_roc_score


def gp_minimize_auc(
    data_dir: str,
    region: Union[Region, str],
    nlo_method: str,
    output_dir: Union[str, os.PathLike] = "_unnamed_optimization",
    n_calls: int = 15,
    esr: Optional[int] = 15,
    random_state: int = 414,
):
    """Minimize AUC using Gaussian processes

    This is our hyperparameter optimization procedure which uses the
    :py:func:`skopt.gp_minimize` functions from Scikit-Optimize.

    Parameters
    ----------
    data_dir : str
       path containing ROOT files
    region : Region or str
       the region where we're going to perform the training
    nlo_method : str
       which tW NLO method sample ('DR' or 'DS' or 'Both')
    output_dir : str or os.PathLike
       path to save optimization output
    n_calls : int
       number of times to train during the minimization procedure
    esr : int, optional
       early stopping rounds for fitting the model
    random_state: int
       random state for splitting data into training/testing

    Examples
    --------

    >>> from tdub.utils import Region
    >>> from tdub.train import prepare_from_root, gp_minimize_auc
    >>> gp_minimize_auc("/path/to/data", Region.r2j1b, "DS", "opt_DS_2j1b")
    >>> gp_minimize_auc("/path/to/data", Region.r2j1b, "DR", "opt_DR_2j1b")

    """

    from skopt.utils import use_named_args
    from skopt.space import Real, Integer
    from skopt.plots import plot_convergence

    from skopt import gp_minimize

    qfiles = quick_files(f"{data_dir}")
    if nlo_method == "DR":
        tW_files = qfiles["tW_DR"]
    elif nlo_method == "DS":
        tW_files = qfiles["tW_DS"]
    elif nlo_method == "Both":
        tW_files = qfiles["tW_DR"] + qfiles["tW_DS"]
        tW_files.sort()
    else:
        raise ValueError("nlo_method must be 'DR' or 'DS' or 'Both'")

    df, labels, weights = prepare_from_root(tW_files, qfiles["ttbar"], region)
    X_train, X_test, y_train, y_test, w_train, w_test = train_test_split(
        df, labels, weights, train_size=0.333, random_state=random_state, shuffle=True
    )
    validation_data = [(X_test, y_test)]
    validation_w = w_test

    n_sig = y_train[y_train == 1].shape[0]
    n_bkg = y_train[y_train == 0].shape[0]
    scale_pos_weight = n_bkg / n_sig
    sample_size = n_bkg + n_sig
    log.info(f"n_bkg / n_sig = {n_bkg} / {n_sig} = {scale_pos_weight}")

    dimensions = [
        Real(low=0.01, high=0.2, prior="log-uniform", name="learning_rate"),
        Integer(low=40, high=250, name="num_leaves"),
        Integer(low=20, high=250, name="min_child_samples"),
        Integer(low=3, high=10, name="max_depth"),
    ]
    default_parameters = [0.1, 100, 50, 5]

    run_from_dir = os.getcwd()
    save_dir = PosixPath(output_dir)
    save_dir.mkdir(exist_ok=True, parents=True)
    os.chdir(save_dir)

    global best_fit
    global best_auc
    global best_parameters
    global best_paramdict
    global ifit
    best_fit = 0
    best_auc = 0.0
    best_parameters = [{"teste": 1}]
    best_paramdict = {}
    ifit = 0

    @use_named_args(dimensions=dimensions)
    def afit(
        learning_rate,
        num_leaves,
        min_child_samples,
        max_depth,
    ):
        global ifit
        global best_fit
        global best_auc
        global best_parameters
        global best_paramdict


        log.info(f"learning_rate: {learning_rate}")
        log.info(f"num_leaves: {num_leaves}")
        log.info(f"min_child_samples: {min_child_samples}")
        log.info(f"max_depth: {max_depth}")

        curdir = os.getcwd()
        p = PosixPath(f"training_{ifit}")
        p.mkdir(exist_ok=False)
        os.chdir(p.resolve())

        with open("features.txt", "w") as f:
            for c in df.columns:
                print(c, file=f)

        model = lgbm.LGBMClassifier(
            boosting_type="gbdt",
            learning_rate=learning_rate,
            num_leaves=num_leaves,
            min_child_samples=min_child_samples,
            max_depth=max_depth,
            n_estimators=500,
            scale_pos_weight=scale_pos_weight,
        )

        fitted_model = model.fit(
            X_train,
            y_train,
            eval_set=validation_data,
            eval_metric="auc",
            verbose=20,
            early_stopping_rounds=esr,
            eval_sample_weight=[validation_w],
        )

        pred = fitted_model.predict_proba(X_test)[:, 1]
        score = roc_auc_score(y_test, pred, sample_weight=w_test)

        train_pred = fitted_model.predict_proba(X_train)[:, 1]
        fig, ax = plt.subplots()
        xmin = np.min(pred[y_test == 0])
        xmax = np.max(pred[y_test == 1])
        bins = np.linspace(0, 1, 41)
        ax.hist(
            train_pred[y_train == 0],
            bins=bins,
            label="Bkg. (train)",
            density=True,
            histtype="step",
            weights=w_train[y_train == 0],
        )
        ax.hist(
            train_pred[y_train == 1],
            bins=bins,
            label="Sig. (train)",
            density=True,
            histtype="step",
            weights=w_train[y_train == 1],
        )
        ax.hist(
            pred[y_test == 0],
            bins=bins,
            label="Bkg. (test)",
            density=True,
            histtype="step",
            weights=w_test[y_test == 0],
        )
        ax.hist(
            pred[y_test == 1],
            bins=bins,
            label="Sig. (test)",
            density=True,
            histtype="step",
            weights=w_test[y_test == 1],
        )
        ax.set_ylim([0, 1.5 * ax.get_ylim()[1]])
        ax.legend(ncol=2, loc="upper center")
        fig.savefig("histograms.pdf")
        plt.close(fig)

        binning_sig_min = min(np.min(pred[y_test == 1]), np.min(train_pred[y_train == 1]))
        binning_sig_max = max(np.max(pred[y_test == 1]), np.max(train_pred[y_train == 1]))
        binning_bkg_min = min(np.min(pred[y_test == 0]), np.min(train_pred[y_train == 0]))
        binning_bkg_max = max(np.max(pred[y_test == 0]), np.max(train_pred[y_train == 0]))
        # binning_sig = np.linspace(binning_sig_min, binning_sig_max, 41)
        # binning_bkg = np.linspace(binning_bkg_min, binning_bkg_max, 41)
        binning_sig = np.linspace(0, 1, 41)
        binning_bkg = np.linspace(0, 1, 41)

        h_sig_test, err_sig_test = pygram11.histogram(
            pred[y_test == 1], bins=binning_sig, weights=w_test[y_test == 1]
        )
        h_sig_train, err_sig_train = pygram11.histogram(
            train_pred[y_train == 1], bins=binning_sig, weights=w_train[y_train == 1]
        )

        h_bkg_test, err_bkg_test = pygram11.histogram(
            pred[y_test == 0], bins=binning_bkg, weights=w_test[y_test == 0]
        )
        h_bkg_train, err_bkg_train = pygram11.histogram(
            train_pred[y_train == 0], bins=binning_bkg, weights=w_train[y_train == 0]
        )

        ks_statistic_sig, ks_pvalue_sig = ks_twosample_binned(
            h_sig_test, h_sig_train, err_sig_test, err_sig_train
        )
        ks_statistic_bkg, ks_pvalue_bkg = ks_twosample_binned(
            h_bkg_test, h_bkg_train, err_bkg_test, err_bkg_train
        )

        if ks_pvalue_sig < 0.1 or ks_pvalue_bkg < 0.1:
            score = score * 0.9

        log.info(f"ksp sig: {ks_pvalue_sig}")
        log.info(f"ksp bkg: {ks_pvalue_bkg}")

        curp = pformat(model.get_params())
        curp = eval(curp)

        with open("params.json", "w") as f:
            json.dump(curp, f, indent=4)

        with open("auc.txt", "w") as f:
            print(f"{score}", file=f)

        with open("ks.txt", "w") as f:
            print(f"sig: {ks_pvalue_sig}", file=f)
            print(f"bkg: {ks_pvalue_bkg}", file=f)

        os.chdir(curdir)

        if score > best_auc:
            best_parameters[0] = model.get_params()
            best_auc = score
            best_fit = ifit
            best_paramdict = curp

        ifit += 1

        del model
        return -score

    search_result = gp_minimize(
        func=afit,
        dimensions=dimensions,
        acq_func="gp_hedge",
        n_calls=n_calls,
        x0=default_parameters,
    )

    summary = {
        "region": region,
        "nlo_method": nlo_method,
        "features": list(df.columns),
        "best_iteration": best_fit,
        "best_auc": best_auc,
        "best_params": best_paramdict,
    }

    with open("summary.json", "w") as f:
        json.dump(summary, f, indent=4)

    fig, ax = plt.subplots()
    plot_convergence(search_result, ax=ax)
    fig.savefig("gpmin_convergence.pdf")

    os.chdir(run_from_dir)
    return 0
